# -*- encoding: binary -*-
# here are internal implementation details, do not use them in your code
require 'stringio'
require 'uri'
require 'mogilefs/chunker'

##
# HTTPFile wraps up the new file operations for storing files onto an HTTP
# storage node.
#
# You really don't want to create an HTTPFile by hand.  Instead you want to
# create a new file using MogileFS::MogileFS.new_file.
#
class MogileFS::HTTPFile < StringIO
  class EmptyResponseError < MogileFS::Error; end
  class BadResponseError < MogileFS::Error; end
  class UnparseableResponseError < MogileFS::Error; end
  class NoStorageNodesError < MogileFS::Error
    def message; 'Unable to open socket to storage node'; end
  end
  class NonRetryableError < MogileFS::Error; end

  class HTTPSock < MogileFS::Socket
    attr_accessor :start

    # Increase timeout as we become more invested in uploading with
    # this socket.  The server could be experiencing I/O delays
    # from large uploads because the sysadmin forgot to tune the
    # VM sysctls for handling large files.
    def write(buf)
      timed_write(buf, Time.now - @start + 5.0)
    end
  end

  # :stopdoc:
  MD5_TRAILER_NODES = {} # :nodoc: # EXPERIMENTAL
  class << self
    attr_accessor :response_timeout_cb
  end

  # temporary directories (nginx) may not be configured on the
  # same device, necessitating a time-consuming full file copy
  # instead of a quick rename(2)/link(2) operation
  @response_timeout_cb = lambda do |elapsed_time, bytes_uploaded|
    mbytes_uploaded = bytes_uploaded / (1024.0 * 1024.0)
    # assumes worst case is 10M/s on the remote storage disk
    t = mbytes_uploaded * 10 + elapsed_time
    t < 5 ? 5 : t
  end
  # :startdoc:

  ##
  # The URI this file will be stored to.

  attr_reader :uri

  attr_reader :devid

  ##
  # The big_io name in case we have file > 256M

  attr_accessor :big_io

  attr_accessor :streaming_io

  ##
  # Creates a new HTTPFile with MogileFS-specific data.  Use
  # MogileFS::MogileFS#new_file instead of this method.

  def initialize(dests, content_length)
    super ""
    @streaming_io = @big_io = @uri = @devid = @active = nil
    @dests = dests
  end

  def request_put(sock, uri, file_size, input = nil)
    host_with_port = "#{uri.host}:#{uri.port}"
    md5 = false
    if MD5_TRAILER_NODES[host_with_port]
      file_size = nil
      md5 = true
    end

    if file_size
      sock.write("PUT #{uri.request_uri} HTTP/1.0\r\n" \
                 "Content-Length: #{file_size}\r\n\r\n")
      input ? MogileFS.io.copy_stream(@active = input, sock) : yield(sock)
    else
      trailers = md5 ? "Trailer: Content-MD5\r\n" : ""
      sock.write("PUT #{uri.request_uri} HTTP/1.1\r\n" \
                 "Host: #{host_with_port}\r\n#{trailers}" \
                 "Transfer-Encoding: chunked\r\n\r\n")
      tmp = MogileFS::Chunker.new(sock, md5)
      rv = input ? MogileFS.io.copy_stream(@active = input, tmp) : yield(tmp)
      tmp.flush
      rv
    end
  end

  def put_streaming_io(sock, uri) # unlikely to be used
    file_size = @streaming_io.length
    written = 0
    request_put(sock, uri, file_size) do |wr|
      @streaming_io.call(Proc.new do |data_to_write|
        written += wr.write(data_to_write)
      end)
    end
    file_size ? file_size : written
  end

  def rewind_or_raise!(uri, err)
    @active.rewind if @active
    rescue => e
      msg = "#{uri} failed with #{err.message} (#{err.class}) and " \
            "retrying is impossible as rewind on " \
            "#{@active.inspect} failed with: #{e.message} (#{e.class})"
      raise NonRetryableError, msg, e.backtrace
  end

  ##
  # Writes an HTTP PUT request to +sock+ to upload the file and
  # returns file size if the socket finished writing
  def upload(devid, uri) # :nodoc:
    start = Time.now
    sock = HTTPSock.tcp(uri.host, uri.port)
    sock.start = start
    file_size = length

    if @streaming_io
      file_size = put_streaming_io(sock, uri)
    elsif @big_io
      if String === @big_io || @big_io.respond_to?(:to_path)
        file = File.open(@big_io)
        stat = file.stat
        file_size = request_put(sock, uri, stat.file? ? stat.size : nil, file)
      else
        size = nil
        if @big_io.respond_to?(:stat)
          stat = @big_io.stat
          size = stat.size if stat.file?
        elsif @big_io.respond_to?(:size)
          size = @big_io.size
        end
        file_size = request_put(sock, uri, size, @big_io)
      end
    else
      rewind
      request_put(sock, uri, file_size, self)
    end

    tout = self.class.response_timeout_cb.call(Time.now - start, file_size)

    case line = sock.timed_read(23, "", tout)
    when %r{^HTTP/\d\.\d\s+(2\d\d)\s} # success!
      file_size
    when nil
      raise EmptyResponseError, 'Unable to read response line from server'
    when %r{^HTTP/\d\.\d\s+(\d+)}
      raise BadResponseError, "HTTP response status from upload: #$1"
    else
      raise UnparseableResponseError,
            "Response line not understood: #{line.inspect}"
    end
    rescue => err
      rewind_or_raise!(uri, err)
      raise
    ensure
      file.close if file
      sock.close if sock
  end

  def commit
    errors = nil
    @dests.each do |devid, path|
      begin
        uri = URI.parse(path)
        bytes_uploaded = upload(devid, uri)
        @devid, @uri = devid, uri
        return bytes_uploaded
      rescue NonRetryableError
        raise
      rescue => e
        errors ||= []
        errors << "#{path} - #{e.message} (#{e.class})"
      end
    end

    raise NoStorageNodesError,
          "all paths failed with PUT: #{errors.join(', ')}", []
  end
end
